
/* debug.c
 * ВК-6 v0.32
 * (c) Васильев И.В., Москва, 16 апреля 2012 г.
 *
 * Программа распространяется под лицензией GNU GPL v3.
 *
 * Отладчик командной строки.
 *
 * При компиляции ВК-6 с графическим интерфейсом данный модуль надо переделать
 * или выборосить, удалив вызов DebugCmd() из exec.c.  Одна из причин --
 * непоследовательное использование OsAlert и прямого вывода на stdout.
 */

#include "cpu.h"
#include "decomp.h"
#include "reg.h"
#include "os.h"
#include "vk6.h"
#include "supp.h"
#include "debug.h"

#include <stdio.h>


static int WrongDbgAdr(void)
/* Выдать сообщение об ошибке и возвратить 0. */
{ OsAlert("debug", "Неверный адрес.");
  flgDisplay = 0;
  return 0;
}


static void WrongDbgCmd(void)
{ OsAlert("debug", "Неизвестная команда отладчика.");
  flgDisplay = 0;
}


static void WrongDbgSyntax(void)
{ OsAlert("debug", "Неправильный синтаксис команды отладчика.");
  flgDisplay = 0;
}


static int GetAdr(const unsigned char **s, int *p)
/* Получить адрес из строки *s, вернуть 0 в случае ошибки. */
{ *p = 0;
  if (!IsDigit(**s)) return WrongDbgAdr();
  do
  { *p = 10*(*p) + (**s) - '0';
    (*s)++;
    if (*p>9999) return WrongDbgAdr();
  } while (IsDigit(**s));
  return 1;
}


static int GetRange(const unsigned char *s, int *iS, int *iE)
/* Считать дипазон (от нуля до двух адресов) из строки *s.
   Вернуть 0 в случае ошибки.
 */
{ Space(&s);
  if (*s)
  { if (GetAdr(&s, iS)==0) return 0;
    Space(&s);
    if (*s)
    { if (GetAdr(&s, iE)==0) return 0;
      Space(&s);
      if ((*s) || (*iE < *iS)) return WrongDbgAdr();
      return 1;
    }
  }
  *iE = -1;
  return 1;
}


static void DebugUn1(int *i)
/* Вывести одну декомпилированную строчку. */
{ int i0 = *i, j;
  const unsigned char *s = Decomp(i, 1);
  printf("%04d ", i0);
  for (j = 0; j<3; j++)
  { int q = i0+j, z = *i;
    if (q>9999) q -= 10000;
    if (z<i0) z += 10000;
    if (i0+j < z) printf("%02x ", bProg[q]);
    else printf("   ");
  }
  PutStringVK(s, 12, 0);
  putchar('\n');
}


static void DebugUn(int *i, int *ik)
/* Команда u отладчика ВК-6. */
{ if (*ik<0)
  { int j = 0;
    do DebugUn1(i);
    while (j++ < 9);
  } else
  do
  { int z = *i;
    DebugUn1(i);
    if (*i <= z) break;
  } while (*i <= *ik);
  flgDisplay = 0;
}


static void DebugDump(int *i0, int *ik)
/* Команда d отладчика ВК-6. */
{ int i, j;
  if (*ik < 0) *ik = *i0 + 49;
  if (*ik > 9999) *ik = 9999;
  for (i = (*i0/10)*10; i <= *ik; i+=10)
  { printf("%04d  ", i);
    for (j = 0; j<10; j++)
    { if ((j+i >= *i0) && (j+i <= *ik)) printf("%02x", bProg[j+i]);
      else printf("  ");
      if (j==4)
      { if (i <= *i0) putchar((i+5 >= *i0)?'-':' ');
        else if (i+10 > *ik) putchar((i+4 <= *ik)?'-':' ');
        else putchar('-');
      } else putchar(' ');
    }
    printf("  ");
    for (j = i; j<i+10; j++)
      if ((j >= *i0) && (j <= *ik)) PutCharVK(bProg[j], '.');
      else printf(" ");
    putchar('\n');
  }
  if (*ik < 9999) *i0 = *ik + 1;
  flgDisplay = 0;
}


static void PrintReg1(real x)
/* Вывод числа с плавающей запятой */
{ int i;
  double n = x;
  static char sz[25];
  sprintf(sz, "%-18.12g", n);
  for (i=0; sz[i]; i++) if (sz[i]=='.') sz[i]=',';
  printf("%s\n", sz);
}


static void DebugReg1(int r)
/* Вывести один регистр. */
{ if (r<1000) printf("R%03d = ", r), PrintReg1(ldR[r]);
  else if (r<8168)
  { int b = bR[r-1000];
    printf("R%d = 0%02xh ('", r, b);
    PutCharVK(b, '.');
    printf("') = %d\n", b);
  } else if ((r>=9000) && (r<=9999))
       puts("Для просмотра регистров функций используйте команду РР ИП.");
  else puts("Такой регистр отсутствует в ВК-6.");
}


static void DumpRegs(int i0, int ik)
/* Вывод байтовых регистров в виде дампа памяти. */
{ int i, j;
  for (i = (i0/10)*10; i<=ik; i+=10)
  { printf("R%d  ", i);
    for (j = 0; j<10; j++)
    { if ((j+i>=i0) && (j+i<=ik)) printf("%02x", bR[j+i-1000]);
      else printf("  ");
      if (j==4)
      { if (i<=i0) putchar((i+5>=i0)?'-':' ');
        else if (i+10>ik) putchar((i+4<=ik)?'-':' ');
        else putchar('-');
      } else putchar(' ');
    }
    printf("  ");
    for (j = i; j<i+10; j++)
      if ((j>=i0) && (j<=ik)) PutCharVK(bR[j-1000], '.');
      else printf(" ");
    putchar('\n');
  }
}


static void DebugRegs(int i0, int ik)
/* Вывести диапазон регистров. */
{ if (ik>i0)
  { int z = ik;
    if (i0<1000) /* Вывести десятичные регистры. */
    { if (z>999) z = 999;
      while (i0<=z)
        DebugReg1(i0++);
    }
    if (ik>999) /* Вывести байтовые данные. */
    { int a = i0;
      if (a<1000) a = 1000;
      if (ik<8168) z = ik; else z = 8167;
      DumpRegs(a, z);
    }
  } else DebugReg1(i0);
  flgDisplay = 0;
}


static void DebugHelp(void)
/* Вывести подсказку отладчика. */
{ puts(" Команды отладчика ВК-6 v" STR_VERSION ":\n\n"
       " .u adr1 adr2  Вывести программу.\n"
       " .d adr1 adr2  Вывести коды программы.\n"
       " .r reg1 reg2  Вывести содержимое регистров.\n"
       " .g =adr adrN  Запустить программу с остановами в adrN.\n"
       " .t =adr cnt   Трассировать внутрь (аналог ПП).\n"
       " .p =adr cnt   Трассировать поверх (step over).\n"
       " .?  Эта подсказка.\n");
  flgDisplay = 0;
}


static void DebugGo(const unsigned char *s)
/* Запустить программу, до 10 точек останова. */
{ if (iMode)
  { puts("Для исполнения программы требуется режим F АВТ.");
    return;
  }
  Space(&s);
  if (*s=='=')
  { int z;
    s++;
    Space(&s);
    if (GetAdr(&s, &z)==0) return;
    iPC = z;
    Space(&s);
  }
  nBreaks = 0;
  while (*s)
  { if (nBreaks>=10)
    { puts("В команде G не может быть больше 10 точек останова.");
      nBreaks = 0;
      return;
    } else if (GetAdr(&s, iBreakAdr+nBreaks)==0)
    { nBreaks = 0;
      return;
    } else nBreaks++;
    Space(&s);
  } /* while (*s) */
  Run(0);
  nBreaks = 0;
}


static void DebugTrace(const unsigned char *s)
/* Трассировать внутрь (step into). */
{ int n = 1;
  if (iMode)
  { puts("Для трассировки программы требуется режим F АВТ.");
    return;
  }
  Space(&s);
  if (*s=='=')
  { int z;
    s++;
    Space(&s);
    if (GetAdr(&s, &z)==0) return;
    iPC = z;
    Space(&s);
  }
  if (*s)
  { if (GetAdr(&s, &n)==0) return;
    if (n==0)
    { puts("Неверное число шагов.");
      return;
    }
    Space(&s);
    if (*s) WrongDbgSyntax();
  }
  while (n-- > 0)
  { Step(0);
    PrintStack();
  }
  flgDisplay = 0;
}


static void DebugPTrace(const unsigned char *s)
/* Трассировать поверх (step over). */
{ int n = 1;
  if (iMode)
  { puts("Для трассировки программы требуется режим F АВТ.");
    return;
  }
  Space(&s);
  if (*s=='=')
  { int z;
    s++;
    Space(&s);
    if (GetAdr(&s, &z)==0) return;
    iPC = z;
    Space(&s);
  }
  if (*s)
  { if (GetAdr(&s, &n)==0) return;
    if (n==0)
    { puts("Неверное число шагов.");
      return;
    }
    Space(&s);
    if (*s) WrongDbgSyntax();
  }
  while (n-- > 0)
  { int b = bProg[iBreakAdr[0]=iPC];
    if ((b>=0xa0) && (b<=0xae)) /* КПП */
    { nBreaks = 1;
      iBreakAdr[0]++;
    } else switch (b)
    { case 0x58: /* FL2 */
      case 0x5a: /* FL3 */
      case 0x5b: /* FL1 */
      case 0x5d: /* FL0 */
      case 0x53: /* ПП */
      case 0xaf: /* РК ПП */
        nBreaks = 1;
        iBreakAdr[0] += 2;
        break;
      case 0xf8: /* PFL2 */
      case 0xfa: /* PFL3 */
      case 0xfb: /* PFL1 */
      case 0xfd: /* PFL0 */
      case 0xf3: /* Р ПП */
        nBreaks = 1;
        iBreakAdr[0] += 3;
        break;
      default:
        nBreaks = 0;
        Step(0);
    }
    if (nBreaks) Run(0), nBreaks = 0;
    PrintStack();
  }
  flgDisplay = 0;
}


void DebugCmd(const unsigned char *s)
/* Исполнить команду отладчика ВК-6. */
{ static int iStart = 0, iRStart = 0;
  int iEnd, iREnd;
  Space(&s);
  if (*s) switch (Up(*s++))
  { case 'D': /* Дамп памяти программ */
      if (GetRange(s, &iStart, &iEnd)) DebugDump(&iStart, &iEnd);
      break;
    case 'G': /* Запуск программы */
      DebugGo(s);
      break;
    case 'P': /* Внешняя трассировка */
      DebugPTrace(s);
      break;
    case 'R': /* Содержимое регистров */
      if (GetRange(s, &iRStart, &iREnd)) DebugRegs(iRStart, iREnd);
      break;
    case 'T': /* Внутренняя трассировка  */
      DebugTrace(s);
      break;
    case 'U': /* Декомпилятор */
      if (GetRange(s, &iStart, &iEnd)) DebugUn(&iStart, &iEnd);
      break;
    case '?': /* Подсказка */
      Space(&s);
      if (*s) WrongDbgCmd(); else DebugHelp();
      break;
    default:
      WrongDbgCmd();
      break;
  } else flgDisplay = 0; /* Разрешить "пустую" команду двоеточие. */
}


