
/* decomp.c
 * ВК-6 v0.31
 * (c) Васильев И.В., Москва, 6 марта 2010 г.
 *
 * Программа распространяется под лицензией GNU GPL v3.
 *
 * Восстановление мнемоники команды ВК-6 по её коду.
 * Учитывает русскую и латинскую мнемоники.
 *
 */

#include "cpu.h"
#include "supp.h"
#include "decomp.h"

int flgLat=0;

unsigned char sMnemo[12];

const char *tblRus00to3f[]=
{ "0", "1", "2", "3", "4", "5", "6", "7",
  "8", "9", ",", "/-/", "\x82\x8f", "Cx", "\x82\xc6", "F \x82x",
  "+", "-", "\xb3", "\xb0", "\xb7", "F 10\xbe", "F e\xbe", "F lg",
  "F ln", "F arcsin", "F arccos", "F arctg", "F sin", "F cos", "F tg", "",
  "F \xb9", "F \xfb", "F x\xbd", "F 1/x",
  "F x\xbf", "F \xb5", "\x8a \x8c\xc5\x83", "\x8a -",
  "\x8a \x88\x8f\x90\x83", "\x8a \xb0", "\x8a \x8c\x91\xc5\x83", "",
  "", "", "", "",
  "\x8a \x83\xc5\x8c\x91", "\x8a |x|", "\x8a \x87\x8d", "\x8a \x83\xc5\x8c",
  "\x8a [x]", "\x8a {x}", "\x8a max", "\x8a AND",
  "\x8a OR", "\x8a XOR", "\x8a NOT", "\x8a \x91\x97",
  "", "", "", "" 
};

const char *tblLat00to3f[]=
{ "0", "1", "2", "3", "4", "5", "6", "7",
  "8", "9", ",", "+/-", "EE", "CX", "ENT", "F ANS",
  "+", "-", "*", "/", "<->", "F 10^X", "F EXP", "F LG",
  "F LN", "F ARCSIN", "F ARCCOS", "F ARCTG", "F SIN", "F COS", "F TG", "",
  "F PI", "F SQRT", "F X^2", "F 1/X", "F X^Y", "F R", "K M->D", "K -",
  "K PRGM", "K /", "K MS->D", "", "", "", "", "",
  "K D->MS", "K ABS", "K SGN", "K D->M", "K INT", "K FRAC", "K MAX", "K AND",
  "K OR", "K XOR", "K NOT", "K RAN", "", "", "", ""
};


int Fetch(int *pc)
{ int c=bProg[(*pc)++];
  if (*pc>9999) *pc = 0;
  return c;
}


const unsigned char *Mnemo1(int c)
/* Получить мнемонику простой одношаговой команды. */
{ if (c<=0x3f) return (unsigned char *)
                      ( (flgLat ? tblLat00to3f : tblRus00to3f) [c] );
  else switch (c)
  { case 0x50: return (unsigned char *)(flgLat? "R/S"   :"\x91/\x8f");
    case 0x52: return (unsigned char *)(flgLat? "RTN"   :"\x82/\x8e");
    case 0x54: return (unsigned char *)(flgLat? "K NOP" :"\x8a \x8d\x8e\x8f");
    case 0x55: return (unsigned char *)(flgLat? "K SCR" :"\x8a \x9d\x8a\x90");
    case 0x56: return (unsigned char *)(flgLat? "K GRPH":"\x8a \x83\x90\x94");
    case 0xf2: return (unsigned char *)(flgLat? "PP RTN":"\x90\x90 \x82/\x8e");
  }
  return (unsigned char *) "";
}


static unsigned char *MnemoR(char *sLat, char *sRus, int r)
/* Собрать мнемонику регистровой команды. */
{ unsigned char *s = (unsigned char *) (flgLat?sLat:sRus);
  int i=0;
  while (s[i]) { sMnemo[i] = s[i]; i++; }
  sMnemo[i++] = ' ';
  sMnemo[i++] = strHex[r];
  sMnemo[i] = '\0';
  return sMnemo;
}


static unsigned char *Mnemo2(char *sLat, char *sRus, int bcd)
/* Собрать мнемонику двухшаговой команды. */
{ unsigned char *s = (unsigned char *) (flgLat?sLat:sRus);
  int i=0;
  while (s[i]) { sMnemo[i] = s[i]; i++; }
  sMnemo[i++] = ' ';
  sMnemo[i++] = strHex[bcd/16];
  sMnemo[i++] = strHex[bcd%16];
  sMnemo[i] = '\0';
  return sMnemo;
}


static unsigned char *Mnemo3(char *sLat, char *sRus, int bcd1, int bcd2)
/* Собрать мнемонику трёхшаговой команды. */
{ unsigned char *s = (unsigned char *) (flgLat?sLat:sRus);
  int i=0;
  while (s[i]) { sMnemo[i] = s[i]; i++; }
  sMnemo[i++] = ' ';
  sMnemo[i++] = strHex[bcd1/16];
  sMnemo[i++] = strHex[bcd1%16];
  sMnemo[i++] = strHex[bcd2/16];
  sMnemo[i++] = strHex[bcd2%16];
  sMnemo[i] = '\0';
  return sMnemo;
}


const unsigned char *MnemoM(int r)
  { return MnemoR("M", "\x8f", r); }

const unsigned char *MnemoRM(int r)
  { return MnemoR("RM", "\x88\x8f", r); }

const unsigned char *MnemoKM(int r)
  { return MnemoR("K M", "\x8a \x8f", r); }

const unsigned char *MnemoKRM(int r)
  { return MnemoR("K RM", "\x8a \x88\x8f", r); }

const unsigned char *MnemoKGOTO(int r)
  { return MnemoR("K GOTO", "\x8a \x81\x8f", r); }

const unsigned char *MnemoKGSB(int r)
  { return MnemoR("K GSB", "\x8a \x8f\x8f", r); }


const unsigned char *MnemoKXNE0(int r)
  { return MnemoR("K X!=0", "\x8a x\xb2" "0", r); }

const unsigned char *MnemoKXGE0(int r)
  { return MnemoR("K X>=0", "\x8a x\xb1" "0", r); }

const unsigned char *MnemoKXLT0(int r)
  { return MnemoR("K X<0", "\x8a x<0", r); }

const unsigned char *MnemoKXEQ0(int r)
  { return MnemoR("K X=0", "\x8a x=0", r); }


const unsigned char *MnemoPM(int bcd)
  { return Mnemo2("P M", "\x90 \x8f", bcd); }

const unsigned char *MnemoPRM(int bcd)
  { return Mnemo2("P RM", "\x90 \x88\x8f", bcd); }

const unsigned char *MnemoPKM(int bcd)
  { return Mnemo2("PK M", "\x90\x8a \x8f", bcd); }

const unsigned char *MnemoPKRM(int bcd)
  { return Mnemo2("PK RM", "\x90\x8a \x88\x8f", bcd); }

const unsigned char *MnemoPPM(int bcd1, int bcd2)
  { return Mnemo3("PP M", "\x90\x90 \x8f", bcd1, bcd2); }

const unsigned char *MnemoPPRM(int bcd1, int bcd2)
  { return Mnemo3("PP RM", "\x90\x90 \x88\x8f", bcd1, bcd2); }

const unsigned char *MnemoGOTO(int bcd)
  { return Mnemo2("GOTO", "\x81\x8f", bcd); }

const unsigned char *MnemoGSB(int bcd)
  { return Mnemo2("GSB", "\x8f\x8f", bcd); }

const unsigned char *MnemoPKGOTO(int bcd)
  { return Mnemo2("PK GOTO", "\x90\x8a \x81\x8f", bcd); }

const unsigned char *MnemoPKGSB(int bcd)
  { return Mnemo2("PK GSB", "\x90\x8a \x8f\x8f", bcd); }

const unsigned char *MnemoPGOTO(int bcd1, int bcd2)
  { return Mnemo3("P GOTO", "\x90 \x81\x8f", bcd1, bcd2); }

const unsigned char *MnemoPGSB(int bcd1, int bcd2)
  { return Mnemo3("P GSB", "\x90 \x8f\x8f", bcd1, bcd2); }


const unsigned char *MnemoFXNE0(int bcd)
  { return Mnemo2("F X!=0", "F x\xb2" "0", bcd); }

const unsigned char *MnemoFXGE0(int bcd)
  { return Mnemo2("F X>=0", "F x\xb1" "0", bcd); }

const unsigned char *MnemoFXLT0(int bcd)
  { return Mnemo2("F X<0", "F x<0", bcd); }

const unsigned char *MnemoFXEQ0(int bcd)
  { return Mnemo2("F X=0", "F x=0", bcd); }


const unsigned char *MnemoPKXNE0(int bcd)
  { return Mnemo2("PK X!=0", "\x90\x8a x\xb2" "0", bcd); }

const unsigned char *MnemoPKXGE0(int bcd)
  { return Mnemo2("PK X>=0", "\x90\x8a x\xb1" "0", bcd); }

const unsigned char *MnemoPKXLT0(int bcd)
  { return Mnemo2("PK X<0", "\x90\x8a x<0", bcd); }

const unsigned char *MnemoPKXEQ0(int bcd)
  { return Mnemo2("PK X=0", "\x90\x8a x=0", bcd); }


const unsigned char *MnemoPXNE0(int bcd1, int bcd2)
  { return Mnemo3("P X!=0", "\x90 x\xb2" "0", bcd1, bcd2); }

const unsigned char *MnemoPXGE0(int bcd1, int bcd2)
  { return Mnemo3("P X>=0", "\x90 x\xb1" "0", bcd1, bcd2); }

const unsigned char *MnemoPXLT0(int bcd1, int bcd2)
  { return Mnemo3("P X<0", "\x90 x<0", bcd1, bcd2); }

const unsigned char *MnemoPXEQ0(int bcd1, int bcd2)
  { return Mnemo3("P X=0", "\x90 x=0", bcd1, bcd2); }


const unsigned char *MnemoFL(int r, int bcd)
{ static char s[] = "F L0";
  s[3] = strHex[r];
  return Mnemo2(s, s, bcd);
}


const unsigned char *MnemoPFL(int r, int bcd1, int bcd2)
{ static char s[] = "PF L0";
  s[0] = flgLat ? 'P' : '\x90' ;
  s[4] = strHex[r];
  return Mnemo3(s, s, bcd1, bcd2);
}


static const unsigned char *Decomp2(int *pc)
{ int c = Fetch(pc);
  if ((c>=0x40) && (c<=0x4e)) return MnemoM(c-0x40);
  else if ((c>=0x60) && (c<=0x6e)) return MnemoRM(c-0x60);
  else if ((c>=0x70) && (c<=0x7e)) return MnemoKXNE0(c-0x70);
  else if ((c>=0x80) && (c<=0x8e)) return MnemoKGOTO(c-0x80);
  else if ((c>=0x90) && (c<=0x9e)) return MnemoKXGE0(c-0x90);
  else if ((c>=0xa0) && (c<=0xae)) return MnemoKGSB(c-0xa0);
  else if ((c>=0xb0) && (c<=0xbe)) return MnemoKM(c-0xb0);
  else if ((c>=0xc0) && (c<=0xce)) return MnemoKXLT0(c-0xc0);
  else if ((c>=0xd0) && (c<=0xde)) return MnemoKRM(c-0xd0);
  else if ((c>=0xe0) && (c<=0xee)) return MnemoKXEQ0(c-0xe0);
  switch (c)
  { case 0x4f: return MnemoPM(Fetch(pc));
    case 0x51: return MnemoGOTO(Fetch(pc));
    case 0x53: return MnemoGSB(Fetch(pc));
    case 0x57: return MnemoFXNE0(Fetch(pc));
    case 0x58: return MnemoFL(2, Fetch(pc));
    case 0x59: return MnemoFXGE0(Fetch(pc));
    case 0x5a: return MnemoFL(3, Fetch(pc));
    case 0x5b: return MnemoFL(1, Fetch(pc));
    case 0x5c: return MnemoFXLT0(Fetch(pc));
    case 0x5d: return MnemoFL(0, Fetch(pc));
    case 0x5e: return MnemoFXEQ0(Fetch(pc));
    case 0x6f: return MnemoPRM(Fetch(pc));
    case 0x7f: return MnemoPKXNE0(Fetch(pc));
    case 0x8f: return MnemoPKGOTO(Fetch(pc));
    case 0x9f: return MnemoPKXGE0(Fetch(pc));
    case 0xaf: return MnemoPKGSB(Fetch(pc));
    case 0xbf: return MnemoPKM(Fetch(pc));
    case 0xcf: return MnemoPKXLT0(Fetch(pc));
    case 0xdf: return MnemoPKRM(Fetch(pc));
    case 0xef: return MnemoPKXEQ0(Fetch(pc));
    case 0xf1: c=Fetch(pc); return MnemoPGOTO(c, Fetch(pc));
    case 0xf3: c=Fetch(pc); return MnemoPGSB(c, Fetch(pc));
    case 0xf4: c=Fetch(pc); return MnemoPPM(c, Fetch(pc));
    case 0xf6: c=Fetch(pc); return MnemoPPRM(c, Fetch(pc));
    case 0xf7: c=Fetch(pc); return MnemoPXNE0(c, Fetch(pc));
    case 0xf8: c=Fetch(pc); return MnemoPFL(2, c, Fetch(pc));
    case 0xf9: c=Fetch(pc); return MnemoPXGE0(c, Fetch(pc));
    case 0xfa: c=Fetch(pc); return MnemoPFL(3, c, Fetch(pc));
    case 0xfb: c=Fetch(pc); return MnemoPFL(1, c, Fetch(pc));
    case 0xfc: c=Fetch(pc); return MnemoPXLT0(c, Fetch(pc));
    case 0xfd: c=Fetch(pc); return MnemoPFL(0, c, Fetch(pc));
    case 0xfe: c=Fetch(pc); return MnemoPXEQ0(c, Fetch(pc));
  }
  return Mnemo1(c);
}


const unsigned char *Decomp(int *pc, int flgMove)
{ int ip=*pc;
  const unsigned char *s=Decomp2(&ip);
  if (flgMove) *pc = ip;
  return s;
}

