
/* exec.c
 * ВК-6 v0.33
 * (c) Васильев И.В., Москва, 5 марта 2010 г.
 *
 * Программа распространяется под лицензией GNU GPL v3.
 *
 * Нажатие клавиш ВК-6, перечисленных в строке.
 *
 * Обратите внимание, что ошибки не меняют строку комментария, а выводятся
 * через системно-зависимый модуль.  Соответственно кодировка этих сообщений
 * должна соответствовать той, которая используется системой.
 *
 * При компиляции ВК-6 с графическим интерфейсом данный модуль надо доработать
 * или выбросить, удалив вызов Execute() из vk6.c.
 */

#include "cpu.h"
#include "exec.h"
#include "clock.h"
#include "keyboard.h"
#include "os.h"
#include "vk6.h"
#include "debug.h"
#include "supp.h"


int flgExit=0;


static void *WrongCmd(void)
/* Выдать сообщение об ошибке и возвратить 0. */
{ flgError=2;
  OsAlert("exec", "Неизвестная команда.");
  return 0;
}


static void *WrongDigit(void)
/* Выдать сообщение об ошибке и возвратить 0. */
{ flgError=2;
  OsAlert("exec", "Неверная цифра.");
  return 0;
}


static void *WrongReg(void)
/* Выдать сообщение об ошибке и возвратить 0. */
{ flgError=2;
  OsAlert("exec", "Неверный регистр.");
  return 0;
}


static int IsCmd(const unsigned char **s, const char *p)
/* Проверить наличие определённой команды во входной строке. */
{ const unsigned char *t = (unsigned char *)p;
  unsigned char c;
  int q = 0;
  while ((c = t[q])!=0) // !=0 against warning
    if (Up((*s)[q++])!=c) return 0;
  (*s) += q;
  return 1;
}


static int IsPRG(const unsigned char **s, const char *p)
/* Команда допустима лишь в режиме F ПРГ. */
{ if (iMode==2) return IsCmd(s, p);
  return 0;
}


static int IsCmd2(const unsigned char **s, const char *p1, const char *p2)
/* Одна команда, две мнемоники. */
{ return IsCmd(s, p1) || IsCmd(s, p2);
}


static int IsPRG2(const unsigned char **s, const char *p1, const char *p2)
{ if (iMode==2) return IsCmd(s, p1) || IsCmd(s, p2);
  return 0;
}


static int IsLetter(const unsigned char **s, const char *p)
/* Проверить на совпадение первый символ из входной строки. */
{ const unsigned char *t = (unsigned char *)p;
  unsigned char c=Up(**s), c1;
  while ((c1 = *t++)!=0) // !=0 against warning
    if (c == c1) { (*s)++; return 1; }
  return 0;
}


static void Press(int c)
/* Нажать и отпустить указанную клавишу. */
{ ButtonPress(c);
  ButtonRelease(c);
}


static const unsigned char *ExecutePlain(const unsigned char *s)
/* Нажать одну кнопку из строки s, вернуть 0 в случае ошибки или Q.
   Префиксные клавиши не нажаты.
 */
{ int c;
  if ((c=*s)!=0) // !=0 against warning
  { s++;
    if (IsDigit(c)) Press(c-'0');
    else switch(Up(c))
    { case ',': Press(KEY_TOCHKA);  break;
      case '-': Press(KEY_MINUS);   break;
      case '*': Press(KEY_MUL);     break;
      case '>': Press(KEY_SFWD);    break;
      case '/':
        if (IsCmd(&s, "-/")) Press(KEY_NEG);  else Press(KEY_DIV);
        break;
      case '+':
        if (IsCmd(&s, "/-")) Press(KEY_NEG);  else Press(KEY_PLUS);
        break;
      case '<':
        if (IsCmd(&s, "->")) Press(KEY_XCHG); else Press(KEY_SBCK);
        break;
      case 0x82: /* В */
      case 'B':
        if (IsLetter(&s, "\x8f"))    Press(KEY_EE); /* П */
        else if (IsLetter(&s, "/") && IsLetter(&s, "0O\x8e")) Press(KEY_RET);
        else return WrongCmd();
        break;
      case 0x91: /* С */
      case 'C':
        if (IsLetter(&s, "X"))        Press(KEY_CX);
        else if (IsCmd(&s, "/\x8f"))  Press(KEY_RS);  /* /П */
        else return WrongCmd();
        break;
      case 'E':
        if (IsLetter(&s, "E"))        Press(KEY_EE);
        else if (IsCmd(&s, "NT"))     Press(KEY_ENTER);
        else return WrongCmd();
        break;
      case 'F': Press(KEY_F);  break;
      case 'G':
        if (IsCmd(&s, "SB"))  Press(KEY_GSB);
        else if (IsCmd(&s, "OTO")) Press(KEY_GOTO);
        else if ((iMode==0) && IsCmd(&s, "RAD")) Press(KEY_GRAD);
        else return WrongCmd();
        break;
      case 0x8a: /* К */
      case 'K': Press(KEY_K);  break;
      case 'L':
        if (iMode && IsCmd(&s, "AT")) Press(KEY_GRAD);
        else return WrongCmd();
        break;
      case 0x8c:
      case 'M': Press(KEY_M);  break;
      case 0x90: /* Р */
        if (iMode && IsCmd(&s, "\x93\x91")) /* УС */
          { Press(KEY_GRAD);  break; }
      case 'P': Press(KEY_P);  break;
      case 'R':
        if (iMode && IsCmd(&s, "US")) Press(KEY_GRAD);
        else if (IsLetter(&s, "M"))   Press(KEY_RM);
        else if (IsCmd(&s, "/S"))     Press(KEY_RS);
        else return WrongCmd();
        break;
      case 0x81: /* Б */
        if (IsLetter(&s, "\x8f")) Press(KEY_GOTO); /* П */
        else return WrongCmd();
        break;
      case 0x88: /* И */
        if (IsLetter(&s, "\x8f")) Press(KEY_RM);   /* П */
        else return WrongCmd();
        break;
      case 0x8b: /* Л */
        if (iMode && IsCmd(&s, "\x80\x92")) Press(KEY_GRAD); /* АТ */
        else return WrongCmd();
        break;
      case 0x8f: /* П */
        if (IsLetter(&s, "\x8f")) Press(KEY_GSB);  /* П */
        else Press(KEY_M);
        break;
      default:  return WrongCmd();
    }
  }
  if (iMode==1) Run(1);
  return s;
}


static const unsigned char *ExecuteF(const unsigned char *s)
/* Нажать одну кнопку из строки s, вернуть 0 в случае ошибки.
   Нажата префиксная клавиша F.
 */
{ if      (IsCmd(&s, "10^X"))    Press(KEY_0);
  else if (IsCmd(&s, "EXP"))     Press(KEY_1);
  else if (IsCmd(&s, "LG"))      Press(KEY_2);
  else if (IsCmd(&s, "LN"))      Press(KEY_3);
  else if (IsCmd(&s, "ARCSIN"))  Press(KEY_4);
  else if (IsCmd(&s, "ARCCOS"))  Press(KEY_5);
  else if (IsCmd(&s, "ARCTG"))   Press(KEY_6);
  else if (IsCmd(&s, "SIN"))     Press(KEY_7);
  else if (IsCmd(&s, "COS"))     Press(KEY_8);
  else if (IsCmd(&s, "TG"))      Press(KEY_9);
  else if (IsLetter(&s, "R"))    Press(KEY_TOCHKA);
  else if (IsCmd2(&s, "AVT", "\x80\x82\x92")) Press(KEY_NEG);   /* АВТ */
  else if (IsCmd2(&s, "PRG", "\x8f\x90\x83")) Press(KEY_EE);    /* ПРГ */
  else if (IsCmd2(&s, "CF" , "\x91" "F"))     Press(KEY_CX);
  else if (IsCmd2(&s, "ANS", "\x82X"))        Press(KEY_ENTER); /* ВX  */
  else if (IsCmd(&s, "PI"))      Press(KEY_PLUS);
  else if (IsCmd(&s, "SQRT"))    Press(KEY_MINUS);
  else if (IsCmd(&s, "X^2"))     Press(KEY_MUL);
  else if (IsCmd(&s, "1/X"))     Press(KEY_DIV);
  else if (IsCmd(&s, "X^Y"))     Press(KEY_XCHG);
  else if (IsPRG(&s, "X<0"))     Press(KEY_SFWD);
  else if (IsPRG(&s, "X=0"))     Press(KEY_SBCK);
  else if (IsPRG(&s, "X>=0"))    Press(KEY_RET);
  else if (IsPRG(&s, "X!=0"))    Press(KEY_RS);
  else if (IsPRG(&s, "L0"))      Press(KEY_RM);
  else if (IsPRG(&s, "L1"))      Press(KEY_M);
  else if (IsPRG(&s, "L2"))      Press(KEY_GOTO);
  else if (IsPRG(&s, "L3"))      Press(KEY_GSB);
  else return WrongCmd();
  return s;
}


static const unsigned char *ExecuteK(const unsigned char *s)
/* Нажать одну кнопку из строки s, вернуть 0 в случае ошибки.
   Нажата префиксная клавиша K.
 */
{ if      (IsCmd2(&s, "NOP", "\x8d\x8e\x8f"))      Press(KEY_0); /* НОП */
  else if (IsCmd2(&s, "SCR", "\x9d\x8a\x90"))      Press(KEY_1); /* ЭКР */
  else if (IsCmd2(&s, "GRPH", "\x83\x90\x94"))     Press(KEY_2); /* ГРФ */
  else if (IsCmd(&s, "D->MS"))                     Press(KEY_3);
  else if (IsCmd2(&s, "ABS", "|X|"))               Press(KEY_4);
  else if (IsCmd2(&s, "SGN", "\x87\x8d"))          Press(KEY_5); /* ЗН  */
  else if (IsCmd(&s, "D->M"))                      Press(KEY_6);
  else if (IsCmd2(&s, "INT", "[X]"))               Press(KEY_7);
  else if (IsCmd2(&s, "FRAC", "{X}"))              Press(KEY_8);
  else if (IsCmd(&s, "MAX"))                       Press(KEY_9);
  else if (IsCmd(&s, "AND"))                       Press(KEY_TOCHKA);
  else if (IsCmd(&s, "OR"))                        Press(KEY_NEG);
  else if (IsCmd(&s, "XOR"))                       Press(KEY_EE);
  else if (IsCmd(&s, "NOT"))                       Press(KEY_CX);
  else if (IsCmd2(&s, "RAN", "\x91\x97"))          Press(KEY_ENTER); /* СЧ */
  else if (IsCmd(&s, "M->D"))                      Press(KEY_PLUS);
  else if (IsCmd2(&s, "PRGM", "\x88\x8f\x90\x83")) Press(KEY_MUL); /* ИПРГ */
  else if (IsCmd(&s, "MS->D"))                     Press(KEY_XCHG);
  else if (IsPRG(&s, "X<0"))                       Press(KEY_SFWD);
  else if (IsPRG(&s, "X=0"))                       Press(KEY_SBCK);
  else if (IsPRG(&s, "X>=0"))                      Press(KEY_RET);
  else if (IsPRG(&s, "X!=0"))                      Press(KEY_RS);
  else if (IsCmd2(&s, "RM", "\x88\x8f"))           Press(KEY_RM);   /* ИП */
  else if (IsPRG2(&s, "GSB", "\x8f\x8f"))          Press(KEY_GSB);  /* ПП */
  else if (IsCmd2(&s, "GOTO", "\x81\x8f"))         Press(KEY_GOTO); /* БП */
  else switch (Up(*s++))
  { case '-':  Press(KEY_MINUS);  break;
    case '/':  Press(KEY_DIV);    break;
    case 'F':  Press(KEY_F);      break;
    case 0x8f: /* П */  
    case 'M':  Press(KEY_M);      break;
    default:   return WrongCmd();
  }
  return s;
}


static const unsigned char *ExecuteP(const unsigned char *s)
/* Нажать одну кнопку из строки s, вернуть 0 в случае ошибки.
   Нажаты префиксные клавиши P или PK.
 */
{ if (IsPRG2(&s, "GSB", "\x8f\x8f"))          Press(KEY_GSB);  /* ПП  */
  else if (IsCmd2(&s, "GOTO", "\x81\x8f"))    Press(KEY_GOTO); /* БП  */
  else if (IsCmd2(&s, "PRG", "\x8f\x90\x83")) Press(KEY_EE);   /* ПРГ */
  else if (IsLetter(&s, "M\x8f"))             Press(KEY_M);    /* П   */
  else if (IsCmd2(&s, "RM", "\x88\x8f"))      Press(KEY_RM);   /* ИП  */
  else if (IsCmd2(&s, "CF", "\x91" "F"))      Press(KEY_CX);   /* СF  */
  else if (IsPRG(&s, "X<0"))                  Press(KEY_SFWD);
  else if (IsPRG(&s, "X=0"))                  Press(KEY_SBCK);
  else if (IsPRG(&s, "X>=0"))                 Press(KEY_RET);
  else if (IsPRG(&s, "X!=0"))                 Press(KEY_RS);
  else if (IsLetter(&s, "F"))                 Press(KEY_F);
  else if (IsLetter(&s, "K\x8a"))             Press(KEY_K);
  else if (IsLetter(&s, "P\x90"))             Press(KEY_P);
  else return WrongCmd();
  return s;
}


static const unsigned char *ExecuteFP3(const unsigned char *s)
/* Нажать одну кнопку из строки s, вернуть 0 в случае ошибки.
   Нажата префиксная клавиша F или P в режиме P PRG.
 */
{ if      (IsCmd2(&s, "PRG", "\x8f\x90\x83")) Press(KEY_EE);   /* ПРГ */
  else if (IsCmd2(&s, "AVT", "\x80\x82\x92")) Press(KEY_NEG);  /* АВТ */
  else if (IsCmd2(&s, "CF", "\x91" "F"))      Press(KEY_CX);   /* СF  */
  else return WrongCmd();
  return s;
}


static const unsigned char *ExecutePF(const unsigned char *s)
/* Нажать одну кнопку из строки s, вернуть 0 в случае ошибки.
   Нажаты префиксные клавиши PF.
 */
{ if (IsPRG(&s, "L0"))                   Press(KEY_RM);
  else if (IsPRG(&s, "L1"))              Press(KEY_M);
  else if (IsPRG(&s, "L2"))              Press(KEY_GOTO);
  else if (IsPRG(&s, "L3"))              Press(KEY_GSB);
  else if (IsCmd2(&s, "CF", "\x91" "F")) Press(KEY_CX); /* СF */
  else return WrongCmd();
  return s;
}


static const unsigned char *ExecutePP(const unsigned char *s)
/* Нажать одну кнопку из строки s, вернуть 0 в случае ошибки.
   Нажаты префиксные клавиши PP.
 */
{ if (IsLetter(&s, "F")) Press(KEY_F);
  else if (IsCmd2(&s, "CF", "\x91" "F")) Press(KEY_CX);  /* СF  */
  else if (IsLetter(&s, "M\x8f"))        Press(KEY_M);   /* П   */
  else if (IsCmd2(&s, "RM", "\x88\x8f")) Press(KEY_RM);  /* ИП  */
  else if ( (iMode==2) && IsLetter(&s, "B\x82") && IsLetter(&s, "/")
            && IsLetter(&s, "0O\x8e") )  Press(KEY_RET); /* В/О */
  else return WrongCmd();
  return s;
}


static const unsigned char *ExecuteDigit(const unsigned char *s)
/* Нажать одну цифру из строки s, вернуть 0 в случае ошибки или Q. */
{ unsigned char c = Up(*s++);
  if (IsDigit(c)) Press(c-'0');
  else if ((c>='A') && (c<='E'))                   Press(c-'A'+10);
  else if ((iMode==3) && (c=='>') && (iNumCnt==0)) Press(KEY_SFWD);
  else if ((iMode==3) && (c=='<') && (iNumCnt==0)) Press(KEY_SBCK);
  else if ((iMode==3) && (c=='P'))                 Press(KEY_P);
  else if ((iKbdState!=7) && (c=='F'))             Press(KEY_MUL);
  else if (iKbdState==7) switch (c)
  { case 0x80:  Press(KEY_TOCHKA);  break;
    case 0x82:  Press(KEY_NEG);     break;
    case 0x91:  Press(KEY_EE);      break;
    case 0x84:  Press(KEY_CX);      break;
    case 0x85:  Press(KEY_ENTER);   break;
    default:    return WrongReg();
  } else return WrongDigit();
  return s;
}


void Execute(const unsigned char *s)
{ Space(&s);
  if (*s=='.') { DebugCmd(s+1);  return; }
  if (flgError == 2) flgError = 0;
  while (s && *s)
  { TimerAdjust();
    if ((*s=='q') || (*s=='Q')) { flgExit = 1;  break; }
    switch (iKbdState)
    { case 0: s = ExecutePlain(s); break;
      case 1: if (iMode==3) s = ExecuteFP3(s); else s = ExecuteF(s);
              break;
      case 2: s = ExecuteK(s);     break;
      case 3: /* Обработка P и PK совмещена. */
              if (iMode==3) { s = ExecuteFP3(s);  break; }
      case 5: s = ExecuteP(s);     break;
      case 4: s = ExecutePF(s);    break;
      case 6: s = ExecutePP(s);    break;
      case 7:
      case 8:
      case 9: s = ExecuteDigit(s); break;
    }
    if (flgError) break;
    Space(&s);
  }
}

